#!/usr/bin/env python3
# type: ignore
from __future__ import annotations

import math

import ibis
import ibis.expr.datatypes as dt
import pytest
from pytest import param

ibis.options.verbose = False


def truncate(number, decimals=0):
    """
    Returns a value truncated to a specific number of decimal places.
    """
    if not isinstance(decimals, int):
        raise TypeError('decimal places must be an integer.')
    elif decimals < 0:
        raise ValueError('decimal places has to be 0 or more.')
    elif decimals == 0:
        return math.trunc(number)

    factor = 10.0 ** decimals
    return math.trunc(number * factor) / factor


#
# OPERATIONS TESTED
#
# Abs
# Add
# Capitalize
# Cast
# Ceil

#
# OPERATIONS REMAINING
#
# Alias
# And
# BitAnd
# BitOr
# BitXor
# Bucket
# CategoryLabel
# Coalesce
# CountDistinct
# Count
# CumulativeMax
# CumulativeMean
# CumulativeMin
# CumulativeSum
# DateAdd
# DateDiff
# DateFromYMD
# DateSub
# DateTruncate
# Date
# DayOfWeekIndex
# DayOfWeekName
# DenseRank
# EndsWith
# Equals
# ExistsSubquery
# Exp
# ExtractDayOfYear
# ExtractDay
# ExtractEpochSeconds
# ExtractHour
# ExtractMillisecond
# ExtractMinute
# ExtractMonth
# ExtractQuarter
# ExtractSecond
# ExtractWeekOfYear
# ExtractYear
# FirstValue
# FloorDivide
# Floor
# GeoArea
# GeoAsBinary
# GeoAsEWKB
# GeoAsEWKT
# GeoAsText
# GeoAzimuth
# GeoBuffer
# GeoCentroid
# GeoContainsProperly
# GeoContains
# GeoCoveredBy
# GeoCovers
# GeoCrosses
# GeoDFullyWithin
# GeoDWithin
# GeoDifference
# GeoDisjoint
# GeoDistance
# GeoEndPoint
# GeoEnvelope
# GeoEquals
# GeoGeometryN
# GeoGeometryType
# GeoIntersection
# GeoIntersects
# GeoIsValid
# GeoLength
# GeoLineLocatePoint
# GeoLineMerge
# GeoLineSubstring
# GeoNPoints
# GeoOrderingEquals
# GeoOverlaps
# GeoPerimeter
# GeoSRID
# GeoSetSRID
# GeoSimplify
# GeoStartPoint
# GeoTouches
# GeoTransform
# GeoUnaryUnion
# GeoUnion
# GeoWithin
# GeoX
# GeoY
# GreaterEqual
# Greater
# Greatest
# GroupConcat
# IdenticalTo
# IfNull
# IntervalFromInteger
# IsNull
# LPad
# LStrip
# Lag
# LastValue
# Lead
# Least
# LessEqual
# Less
# Ln
# Log10
# Log2
# Log
# Lowercase
# Max
# Mean
# MinRank
# Min
# Modulus
# Multiply
# NTile
# Negate
# NotAll
# NotAny
# NotContains
# NotEquals
# NotNull
# Not
# NullIfZero
# NullIf
# NullLiteral
# Or
# PercentRank
# Power
# RPad
# RStrip
# RandomScalar
# RegexSearch
# Repeat
# Reverse
# Round
# RowNumber
# SearchedCase
# Sign
# SimpleCase
# Sqrt
# StandardDev
# StartsWith
# StrRight
# Strftime
# StringAscii
# StringConcat
# StringContains
# StringFind
# StringJoin
# StringLength
# StringReplace
# StringSQLILike
# Strip
# Substring
# Subtract
# Sum
# TableArrayView
# TableColumn
# TimeFromHMS
# TimestampAdd
# TimestampDiff
# TimestampFromYMDHMS
# TimestampNow
# TimestampSub
# TimestampTruncate
# TypeOf
# Uppercase
# ValueList
# Variance
# WindowOp

OPERATIONS = [
    # Testing Abs
    ('abs()', 'negative_int_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'int_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'float_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'negative_float_c', lambda x: x.abs(), lambda x: abs(x)),
    # ("abs()", "bool_c", lambda x : x.abs(), lambda x : abs(x)),
    ('abs()', 'tinyint_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'smallint_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'mediumint_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'bigint_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'double_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'decimal_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'decimal5_c', lambda x: x.abs(), lambda x: abs(x)),
    ('abs()', 'numeric_c', lambda x: x.abs(), lambda x: abs(x)),

    # Abs should produce an error with the following types:
    ('error', 'date_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'time_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'time6_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'datetime_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'datetime6_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'timestamp_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'timestamp6_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'char32_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'varchar42_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'longtext_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'mediumtext_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'tinytext_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'text_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'text4_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'blob_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'enumsml_c', lambda x: x.abs(), lambda x: abs(x)),
    ('error', 'set_abcd_c', lambda x: x.abs(), lambda x: abs(x)),

    # Testing Add
    ('add()', 'int_c', lambda x: x.add(3), lambda x: x + 3),
    ('add()', 'int_c', lambda x: x.add(-5), lambda x: x - 5),
    ('add()', 'negative_int_c', lambda x: x.add(3), lambda x: x + 3),
    ('add()', 'negative_int_c', lambda x: x.add(-5), lambda x: x - 5),
    ('add()', 'float_c', lambda x: x.add(3), lambda x: x + 3),
    ('add()', 'float_c', lambda x: x.add(-5), lambda x: x - 5),
    ('add()', 'float_c', lambda x: x.add(-5.3), lambda x: x - 5.3),
    # ("add()", "bool_c", lambda x : x.add(-3), lambda x : x - 3),
    ('add()', 'tinyint_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'smallint_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'mediumint_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'bigint_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'double_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'decimal_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'decimal5_c', lambda x: x.add(-3), lambda x: x - 3),
    ('add()', 'numeric_c', lambda x: x.add(-3), lambda x: x - 3),

    # Add should produce an error with the following types:
    ('error', 'date_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'time_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'time6_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'datetime_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'datetime6_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'timestamp_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'timestamp6_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'char32_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'varchar42_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'longtext_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'mediumtext_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'tinytext_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'text_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'text4_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'blob_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'enumsml_c', lambda x: x.add(-3), lambda x: x - 3),
    ('error', 'set_abcd_c', lambda x: x.add(-3), lambda x: x - 3),

    # Testing And
    # ("And()", "bool_c", lambda x: x.And(True), lambda x: x and True),
    ('error', 'int_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'int_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'negative_int_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'negative_int_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'float_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'float_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'float_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'tinyint_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'smallint_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'mediumint_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'bigint_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'double_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'decimal_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'decimal5_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'numeric_c', lambda x: x.And(True), lambda x: x and True),

    # And should produce an error with the following types:
    ('error', 'date_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'time_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'time6_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'datetime_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'datetime6_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'timestamp_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'timestamp6_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'char32_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'varchar42_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'longtext_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'mediumtext_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'tinytext_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'text_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'text4_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'blob_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'enumsml_c', lambda x: x.And(True), lambda x: x and True),
    ('error', 'set_abcd_c', lambda x: x.And(True), lambda x: x and True),

    # Testing Capitalize
    ('capitalize()', 'char32_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'varchar42_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'longtext_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'mediumtext_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'tinytext_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'text_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('capitalize()', 'text4_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    # ("capitalize()", "set_abcd_c", lambda x: x.capitalize(),
    #                                lambda x: list([i.capitalize() for i in x])),

    # Capitalize should produce an error with the following types:
    ('error', 'negative_int_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'int_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'float_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'negative_float_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    # ("error", "bool_c", lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'tinyint_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'smallint_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'mediumint_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'bigint_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'double_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'decimal_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'decimal5_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'numeric_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'date_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'time_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'time6_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'datetime_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'datetime6_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'timestamp_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'timestamp6_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'blob_c', lambda x: x.capitalize(), lambda x: x.capitalize()),
    ('error', 'enumsml_c', lambda x: x.capitalize(), lambda x: x.capitalize()),

    # Testing Cast
    ('cast()', 'negative_int_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'int_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'float_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'negative_float_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "bool_c", lambda x : x.cast(dt.string), lambda x : str(x)),
    ('cast()', 'tinyint_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'smallint_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'mediumint_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'bigint_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "double_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "decimal_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "decimal5_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "numeric_c", lambda x: x.cast(dt.string), lambda x: str(x)),

    # Cast should produce an error with the following types:
    # ("cast()", "date_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "time_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "time6_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'datetime_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "datetime6_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'timestamp_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "timestamp6_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'char32_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'varchar42_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'longtext_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'mediumtext_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'tinytext_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'text_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    ('cast()', 'text4_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "blob_c", lambda x: x.cast(dt.string), lambda x: str(x)),
    ('error',  'enumsml_c', lambda x: x.cast(dt.string), lambda x: str(x)),
    # ("cast()", "set_abcd_c", lambda x: x.cast(dt.string), lambda x: str(x)),

    # Testing Ceil
    ('ceil()', 'negative_int_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'int_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'float_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'negative_float_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    # ("ceil()", "bool_c", lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'tinyint_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'smallint_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'mediumint_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'bigint_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'double_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'decimal_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'decimal5_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('ceil()', 'numeric_c', lambda x: x.ceil(), lambda x: math.ceil(x)),

    # Ceil should produce an error with the following types:
    ('error', 'date_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'time_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'time6_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'datetime_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'datetime6_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'timestamp_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'timestamp6_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'char32_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'varchar42_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'longtext_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'mediumtext_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'tinytext_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'text_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'text4_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'blob_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'enumsml_c', lambda x: x.ceil(), lambda x: math.ceil(x)),
    ('error', 'set_abcd_c', lambda x: x.ceil(), lambda x: math.ceil(x)),

    ('log(10)', 'int_c', lambda x: x.log(10), lambda x: math.log(x, 10)),
    ('log(2)', 'int_c', lambda x: x.log(2), lambda x: math.log(x, 2)),
    ('log10()', 'int_c', lambda x: x.log10(), lambda x: math.log10(x)),
    ('round()', 'float_c', lambda x: x.round(), lambda x: round(x)),
]


@pytest.mark.parametrize(
    ('test_name', 'column', 'ibis_operation', 'expected_operation'),
    [
        param(test_name, column, ibis_operation, expected_operation, id=test_name)
        for test_name, column, ibis_operation, expected_operation in OPERATIONS
    ],
)
def test_operations(con, test_name, column, ibis_operation, expected_operation):
    datatypes = con.table('datatypes')
    if test_name == 'error':
        try:
            out = datatypes[
                getattr(datatypes, column),
                ibis_operation(getattr(datatypes, column)).name('_'),
            ].execute()
            original = list(out.iloc[:, 0])
            result = list(out.iloc[:, 1])
        except AttributeError:
            return
        assert False

    out = datatypes[
        getattr(datatypes, column),
        ibis_operation(getattr(datatypes, column)).name('_'),
    ].execute()
    original = list(out.iloc[:, 0])
    result = list(out.iloc[:, 1])

    for i, value in enumerate(original):
        ibis_result = result[i]
        expected_result = expected_operation(value)

        if test_name == 'abs()' and type(ibis_result) == float:
            ibis_result = truncate(ibis_result, 2)
            expected_result = truncate(expected_result, 2)

        elif test_name == 'add()' and type(ibis_result) == float:
            if column == 'decimal5_c':
                ibis_result = round(ibis_result, 3)
                expected_result = round(expected_result, 3)
            else:
                ibis_result = truncate(ibis_result, 3)
                expected_result = truncate(expected_result, 3)

        elif test_name == 'cast()' and type(ibis_result) == str \
                and ('float_c' in column or 'timestamp_c' in column):
            max_len = min(len(ibis_result), len(expected_result))
            ibis_result = ibis_result[:max_len-1]
            expected_result = expected_result[:max_len-1]

        assert ibis_result == expected_result



from __future__ import annotations

import json
from typing import Iterable

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class DotProduct(Value):
    """32-bit float dot product."""

    left = binary
    right = binary

    output_dtype = dt.float64
    output_shape = rlz.shape_like('left')


def dot_product(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 32-bit float vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack()
    return DotProduct(left, right).to_expr()


ir.BinaryValue.dot_product = dot_product


class DotProductI8(DotProduct):
    """8-bit integer dot product."""


def dot_product_i8(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 8-bit int vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i8()
    return DotProductI8(left, right).to_expr()


ir.BinaryValue.dot_product_i8 = dot_product_i8


class DotProductI16(DotProduct):
    """16-bit integer dot product."""


def dot_product_i16(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 16-bit int vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i16()
    return DotProductI16(left, right).to_expr()


ir.BinaryValue.dot_product_i16 = dot_product_i16


class DotProductI32(DotProduct):
    """32-bit integer dot product."""


def dot_product_i32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 32-bit int vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i32()
    return DotProductI32(left, right).to_expr()


ir.BinaryValue.dot_product_i32 = dot_product_i32


class DotProductI64(DotProduct):
    """64-bit integer dot product."""


def dot_product_i64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 64-bit int vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i64()
    return DotProductI64(left, right).to_expr()


ir.BinaryValue.dot_product_i64 = dot_product_i64


class DotProductF32(DotProduct):
    """32-bit float dot product."""


def dot_product_f32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 32-bit float vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f32()
    return DotProductF32(left, right).to_expr()


ir.BinaryValue.dot_product_f32 = dot_product_f32


class DotProductF64(DotProduct):
    """64-bit float dot product."""


def dot_product_f64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.FloatingValue:
    """
    Return the scalar product, or dot product of two 64-bit float vector input values.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Double column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f64()
    return DotProductF64(left, right).to_expr()


ir.BinaryValue.dot_product_f64 = dot_product_f64



from __future__ import annotations

import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONDeleteKey(Value):
    """JSON delete key."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]), min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_delete_key(
    arg: ir.JSONValue | ir.StringValue,
    *key_path: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Delete a key from a JSON map or array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Keys or zero-indexed array positions.

    Returns
    -------
    JSON or string column

    """
    return JSONDeleteKey(arg, key_path).to_expr()


ir.JSONValue.json_delete_key = json_delete_key
ir.StringValue.json_delete_key = json_delete_key



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorKthElement(Value):
    """32-bit float kth element."""

    arg = binary
    n = rlz.integer

    output_dtype = dt.float32
    output_shape = rlz.shape_like('arg')


def vector_kth_element(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.FloatingValue:
    """
    Return the 32-bit float kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    32-bit float column

    """
    return VectorKthElement(arg, n).to_expr()


ir.BinaryValue.vector_kth_element = vector_kth_element


class VectorKthElementI8(VectorKthElement):
    """8-bit integer kth element."""

    output_dtype = dt.int8


def vector_kth_element_i8(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.IntegerValue:
    """
    Return the 8-bit int kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    8-bit int column

    """
    return VectorKthElementI8(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_i8 = vector_kth_element_i8


class VectorKthElementI16(VectorKthElement):
    """16-bit integer kth element."""

    output_dtype = dt.int16


def vector_kth_element_i16(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.IntegerValue:
    """
    Return the 16-bit int kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    16-bit int column

    """
    return VectorKthElementI16(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_i16 = vector_kth_element_i16


class VectorKthElementI32(VectorKthElement):
    """32-bit integer kth element."""

    output_dtype = dt.int32


def vector_kth_element_i32(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.IntegerValue:
    """
    Return the 32-bit int kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    32-bit int column

    """
    return VectorKthElementI32(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_i32 = vector_kth_element_i32


class VectorKthElementI64(VectorKthElement):
    """64-bit integer kth element."""

    output_dtype = dt.int64


def vector_kth_element_i64(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.IntegerValue:
    """
    Return the 64-bit int kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    64-bit int column

    """
    return VectorKthElementI64(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_i64 = vector_kth_element_i64


class VectorKthElementF32(VectorKthElement):
    """32-bit float kth element."""

    output_dtype = dt.float32


def vector_kth_element_f32(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.FloatingValue:
    """
    Return the 32-bit float kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    32-bit float column

    """
    return VectorKthElementF32(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_f32 = vector_kth_element_f32


class VectorKthElementF64(VectorKthElement):
    """64-bit float kth element."""

    output_dtype = dt.float64


def vector_kth_element_f64(arg: ir.BinaryValue, n: ir.IntegerValue) -> ir.FloatingValue:
    """
    Return the 64-bit float kth element in a zero-indexed vector expression.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int
        Element index

    Returns
    -------
    64-bit float column

    """
    return VectorKthElementF64(arg, n).to_expr()


ir.BinaryValue.vector_kth_element_f64 = vector_kth_element_f64



from __future__ import annotations

import json
from typing import Iterable

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorMul(Value):
    """32-bit float vector multiplication."""

    left = binary
    right = binary

    output_dtype = dt.binary
    output_shape = rlz.shape_like('left')


def vector_mul(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 32-bit float vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack()
    return VectorMul(left, right).to_expr()


ir.BinaryValue.vector_mul = vector_mul


class VectorMulI8(VectorMul):
    """8-bit integer vector multiplication."""


def vector_mul_i8(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 8-bit int vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i8()
    return VectorMulI8(left, right).to_expr()


ir.BinaryValue.vector_mul_i8 = vector_mul_i8


class VectorMulI16(VectorMul):
    """16-bit integer vector multiplication."""


def vector_mul_i16(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 16-bit int vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i16()
    return VectorMulI16(left, right).to_expr()


ir.BinaryValue.vector_mul_i16 = vector_mul_i16


class VectorMulI32(VectorMul):
    """32-bit integer vector multiplication."""


def vector_mul_i32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 32-bit int vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i32()
    return VectorMulI32(left, right).to_expr()


ir.BinaryValue.vector_mul_i32 = vector_mul_i32


class VectorMulI64(VectorMul):
    """64-bit integer vector multiplication."""


def vector_mul_i64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 64-bit int vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i64()
    return VectorMulI64(left, right).to_expr()


ir.BinaryValue.vector_mul_i64 = vector_mul_i64


class VectorMulF32(VectorMul):
    """32-bit float vector multiplication."""


def vector_mul_f32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 32-bit float vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f32()
    return VectorMulF32(left, right).to_expr()


ir.BinaryValue.vector_mul_f32 = vector_mul_f32


class VectorMulF64(VectorMul):
    """64-bit float vector multiplication."""


def vector_mul_f64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Multiplies two 64-bit float vector blobs.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f64()
    return VectorMulF64(left, right).to_expr()


ir.BinaryValue.vector_mul_f64 = vector_mul_f64



from __future__ import annotations

import json
from typing import Iterable

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorSub(Value):
    """32-bit float vector subtraction."""

    left = binary
    right = binary

    output_dtype = dt.binary
    output_shape = rlz.shape_like('left')


def vector_sub(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 32-bit float vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack()
    return VectorSub(left, right).to_expr()


ir.BinaryValue.vector_sub = vector_sub


class VectorSubI8(VectorSub):
    """8-bit integer vector subtraction."""


def vector_sub_i8(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 8-bit int vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_i8()
    return VectorSubI8(left, right).to_expr()


ir.BinaryValue.vector_sub_i8 = vector_sub_i8


class VectorSubI16(VectorSub):
    """16-bit integer vector subtraction."""


def vector_sub_i16(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 16-bit int vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_i16()
    return VectorSubI16(left, right).to_expr()


ir.BinaryValue.vector_sub_i16 = vector_sub_i16


class VectorSubI32(VectorSub):
    """32-bit integer vector subtraction."""


def vector_sub_i32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 32-bit int vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_i32()
    return VectorSubI32(left, right).to_expr()


ir.BinaryValue.vector_sub_i32 = vector_sub_i32


class VectorSubI64(VectorSub):
    """64-bit integer vector subtraction."""


def vector_sub_i64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 64-bit int vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_i64()
    return VectorSubI64(left, right).to_expr()


ir.BinaryValue.vector_sub_i64 = vector_sub_i64


class VectorSubF32(VectorSub):
    """32-bit float vector subtraction."""


def vector_sub_f32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 32-bit float vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_f32()
    return VectorSubF32(left, right).to_expr()


ir.BinaryValue.vector_sub_f32 = vector_sub_f32


class VectorSubF64(VectorSub):
    """64-bit float vector subtraction."""


def vector_sub_f64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Subtract the second 64-bit float vector from the first.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(right)).json_array_pack_f64()
    return VectorSubF64(left, right).to_expr()


ir.BinaryValue.vector_sub_f64 = vector_sub_f64



from __future__ import annotations

from typing import Any

import ibis
import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest

ROW_IDS = [21, 25, 31]
FLOAT_VECTORS = [
    [0.1, 2.5, -6.7],
    [10.3, -3.3, 7.8],
    [-2.3, 6.6, 7.9],
]
FLOAT32_SERIES = pd.Series(FLOAT_VECTORS, dtype='<f4')
FLOAT64_SERIES = pd.Series(FLOAT_VECTORS, dtype='<f8')
FLOAT32_VECTOR_BYTES = [np.array(x, dtype='<f4').tobytes() for x in FLOAT_VECTORS]
FLOAT64_VECTOR_BYTES = [np.array(x, dtype='<f8').tobytes() for x in FLOAT_VECTORS]
INT_VECTORS = [
    [0, 3, -7],
    [10, -3, 8],
    [-2, 7, 8],
]
INT8_SERIES = pd.Series(INT_VECTORS, dtype='<i1')
INT16_SERIES = pd.Series(INT_VECTORS, dtype='<i2')
INT32_SERIES = pd.Series(INT_VECTORS, dtype='<i4')
INT64_SERIES = pd.Series(INT_VECTORS, dtype='<i8')
INT8_VECTOR_BYTES = [np.array(x, dtype='<i1').tobytes() for x in INT_VECTORS]
INT16_VECTOR_BYTES = [np.array(x, dtype='<i2').tobytes() for x in INT_VECTORS]
INT32_VECTOR_BYTES = [np.array(x, dtype='<i4').tobytes() for x in INT_VECTORS]
INT64_VECTOR_BYTES = [np.array(x, dtype='<i8').tobytes() for x in INT_VECTORS]
LISTS = [
    [0.1, 'foo', True],
    [10.3, 'bar', False],
    [0.3, 'baz', False],
]
OBJECTS = [
    {'a': 1, 'b': 2, 'c': {'d': 100}},
    {'a': 1, 'b': 12, 'c': {'d': 105, 'e': True}},
    {'a': 10, 'b': 25, 'c': {'d': 111, 'e': 1.234}},
]


def make_vectors(tbl: Any) -> Any:
    return tbl.mutate(
        vec1=lambda x: x.text_vector.json_array_pack(),
        vec2=lambda x: x.text_vector.json_array_pack(),
        vec1_f32=lambda x: x.text_vector.json_array_pack_f32(),
        vec2_f32=lambda x: x.text_vector.json_array_pack_f32(),
        vec1_f64=lambda x: x.text_vector.json_array_pack_f64(),
        vec2_f64=lambda x: x.text_vector.json_array_pack_f64(),
        vec1_i8=lambda x: x.text_vector.json_array_pack_i8(),
        vec2_i8=lambda x: x.text_vector.json_array_pack_i8(),
        vec1_i16=lambda x: x.text_vector.json_array_pack_i16(),
        vec2_i16=lambda x: x.text_vector.json_array_pack_i16(),
        vec1_i32=lambda x: x.text_vector.json_array_pack_i32(),
        vec2_i32=lambda x: x.text_vector.json_array_pack_i32(),
        vec1_i64=lambda x: x.text_vector.json_array_pack_i64(),
        vec2_i64=lambda x: x.text_vector.json_array_pack_i64(),
    )


def test_json_array_pack(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        # default (float32)
        text_packed=lambda x: x.text_vector.json_array_pack(),
        json_packed=lambda x: x.json_vector.json_array_pack(),
        text_unpacked=lambda x: x.text_vector.json_array_pack().json_array_unpack(),
        json_unpacked=lambda x: x.json_vector.json_array_pack().json_array_unpack(),

        # float32
        text_float32_packed=lambda x: x.text_vector.json_array_pack_f32(),
        json_float32_packed=lambda x: x.json_vector.json_array_pack_f32(),
        text_float32_unpacked=lambda x: x.text_vector.json_array_pack_f32()\
                                                     .json_array_unpack_f32(),
        json_float32_unpacked=lambda x: x.json_vector.json_array_pack_f32()\
                                                     .json_array_unpack_f32(),

        # float64
        text_float64_packed=lambda x: x.text_vector.json_array_pack_f64(),
        json_float64_packed=lambda x: x.json_vector.json_array_pack_f64(),
        text_float64_unpacked=lambda x: x.text_vector.json_array_pack_f64()\
                                                     .json_array_unpack_f64(),
        json_float64_unpacked=lambda x: x.json_vector.json_array_pack_f64()\
                                                     .json_array_unpack_f64(),

        # int8
        text_int8_packed=lambda x: x.text_vector.json_array_pack_i8(),
        json_int8_packed=lambda x: x.json_vector.json_array_pack_i8(),
        text_int8_unpacked=lambda x: x.text_vector.json_array_pack_i8()\
                                                  .json_array_unpack_i8(),
        json_int8_unpacked=lambda x: x.json_vector.json_array_pack_i8()\
                                                  .json_array_unpack_i8(),

        # int16
        text_int16_packed=lambda x: x.text_vector.json_array_pack_i16(),
        json_int16_packed=lambda x: x.json_vector.json_array_pack_i16(),
        text_int16_unpacked=lambda x: x.text_vector.json_array_pack_i16()\
                                                   .json_array_unpack_i16(),
        json_int16_unpacked=lambda x: x.json_vector.json_array_pack_i16()\
                                                   .json_array_unpack_i16(),

        # int32
        text_int32_packed=lambda x: x.text_vector.json_array_pack_i32(),
        json_int32_packed=lambda x: x.json_vector.json_array_pack_i32(),
        text_int32_unpacked=lambda x: x.text_vector.json_array_pack_i32()\
                                                   .json_array_unpack_i32(),
        json_int32_unpacked=lambda x: x.json_vector.json_array_pack_i32()\
                                                   .json_array_unpack_i32(),

        # int64
        text_int64_packed=lambda x: x.text_vector.json_array_pack_i64(),
        json_int64_packed=lambda x: x.json_vector.json_array_pack_i64(),
        text_int64_unpacked=lambda x: x.text_vector.json_array_pack_i64()\
                                                   .json_array_unpack_i64(),
        json_int64_unpacked=lambda x: x.json_vector.json_array_pack_i64()\
                                                   .json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_packed.tolist() == FLOAT32_VECTOR_BYTES
    assert out.json_packed.tolist() == FLOAT32_VECTOR_BYTES
    tm.assert_series_equal(out.text_unpacked, FLOAT32_SERIES, check_names=False)
    tm.assert_series_equal(out.json_unpacked, FLOAT32_SERIES, check_names=False)

    assert out.text_float32_packed.tolist() == FLOAT32_VECTOR_BYTES
    assert out.json_float32_packed.tolist() == FLOAT32_VECTOR_BYTES
    tm.assert_series_equal(out.text_float32_unpacked, FLOAT32_SERIES, check_names=False)
    tm.assert_series_equal(out.json_float32_unpacked, FLOAT32_SERIES, check_names=False)

    assert out.text_float64_packed.tolist() == FLOAT64_VECTOR_BYTES
    assert out.json_float64_packed.tolist() == FLOAT64_VECTOR_BYTES
    tm.assert_series_equal(out.text_float64_unpacked, FLOAT64_SERIES, check_names=False)
    tm.assert_series_equal(out.json_float64_unpacked, FLOAT64_SERIES, check_names=False)

    assert out.text_int8_packed.tolist() == INT8_VECTOR_BYTES
    assert out.json_int8_packed.tolist() == INT8_VECTOR_BYTES
    tm.assert_series_equal(out.text_int8_unpacked, INT8_SERIES, check_names=False)
    tm.assert_series_equal(out.json_int8_unpacked, INT8_SERIES, check_names=False)

    assert out.text_int16_packed.tolist() == INT16_VECTOR_BYTES
    assert out.json_int16_packed.tolist() == INT16_VECTOR_BYTES
    tm.assert_series_equal(out.text_int16_unpacked, INT16_SERIES, check_names=False)
    tm.assert_series_equal(out.json_int16_unpacked, INT16_SERIES, check_names=False)

    assert out.text_int32_packed.tolist() == INT32_VECTOR_BYTES
    assert out.json_int32_packed.tolist() == INT32_VECTOR_BYTES
    tm.assert_series_equal(out.text_int32_unpacked, INT32_SERIES, check_names=False)
    tm.assert_series_equal(out.json_int32_unpacked, INT32_SERIES, check_names=False)

    assert out.text_int64_packed.tolist() == INT64_VECTOR_BYTES
    assert out.json_int64_packed.tolist() == INT64_VECTOR_BYTES
    tm.assert_series_equal(out.text_int64_unpacked, INT64_SERIES, check_names=False)
    tm.assert_series_equal(out.json_int64_unpacked, INT64_SERIES, check_names=False)


def test_json_array_contains(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_contains_double=lambda x: x.text_vector.json_array_contains_double(0.1),
        text_contains_string=lambda x: x.text_list.json_array_contains_string('bar'),
        text_contains_json=lambda x: x.text_list.json_array_contains_json('true'),
        text_contains=lambda x: x.text_list.json_array_contains(True),

        json_contains_double=lambda x: x.json_vector.json_array_contains_double(0.1),
        json_contains_string=lambda x: x.json_list.json_array_contains_string('bar'),
        json_contains_json=lambda x: x.json_list.json_array_contains_json('true'),
        json_contains=lambda x: x.json_list.json_array_contains(True),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_contains_double.tolist() == [True, False, False]
    assert out.text_contains_string.tolist() == [False, True, False]
    assert out.text_contains_json.tolist() == [True, False, False]
    assert out.text_contains.tolist() == [True, False, False]

    assert out.json_contains_double.tolist() == [True, False, False]
    assert out.json_contains_string.tolist() == [False, True, False]
    assert out.json_contains_json.tolist() == [True, False, False]
    assert out.json_contains.tolist() == [True, False, False]


def test_json_array_push(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_vector_push_double=lambda x: x.text_vector.json_array_push_double(5.5),
        text_list_push_double=lambda x: x.text_list.json_array_push_double(5.5),
        text_obj_push_double=lambda x: x.text_obj.json_array_push_double(5.5),

        json_vector_push_double=lambda x: x.json_vector.json_array_push_double(5.5),
        json_list_push_double=lambda x: x.json_list.json_array_push_double(5.5),
        json_obj_push_double=lambda x: x.json_obj.json_array_push_double(5.5),

        text_vector_push_string=lambda x: x.text_vector.json_array_push_string('hi'),
        text_list_push_string=lambda x: x.text_list.json_array_push_string('hi'),
        text_obj_push_string=lambda x: x.text_obj.json_array_push_string('hi'),

        json_vector_push_string=lambda x: x.json_vector.json_array_push_string('hi'),
        json_list_push_string=lambda x: x.json_list.json_array_push_string('hi'),
        json_obj_push_string=lambda x: x.json_obj.json_array_push_string('hi'),

        text_vector_push_json=lambda x: x.text_vector.json_array_push_json('{"x": true}'),
        text_list_push_json=lambda x: x.text_list.json_array_push_json('{"x": true}'),
        text_obj_push_json=lambda x: x.text_obj.json_array_push_json('{"x": true}'),

        json_vector_push_json=lambda x: x.json_vector.json_array_push_json('{"x": true}'),
        json_list_push_json=lambda x: x.json_list.json_array_push_json('{"x": true}'),
        json_obj_push_json=lambda x: x.json_obj.json_array_push_json('{"x": true}'),

        json_vector_push=lambda x: x.json_vector.json_array_push({'x': True}),
        json_list_push=lambda x: x.json_list.json_array_push({'x': True}),
        json_obj_push=lambda x: x.json_obj.json_array_push({'x': True}),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_vector_push_double.tolist() == [x + [5.5] for x in FLOAT_VECTORS]
    assert out.text_list_push_double.tolist() == [x + [5.5] for x in LISTS]
    assert out.text_obj_push_double.tolist() == [None, None, None]

    assert out.json_vector_push_double.tolist() == [x + [5.5] for x in FLOAT_VECTORS]
    assert out.json_list_push_double.tolist() == [x + [5.5] for x in LISTS]
    assert out.json_obj_push_double.tolist() == [None, None, None]

    assert out.text_vector_push_string.tolist() == [x + ['hi'] for x in FLOAT_VECTORS]
    assert out.text_list_push_string.tolist() == [x + ['hi'] for x in LISTS]
    assert out.text_obj_push_string.tolist() == [None, None, None]

    assert out.json_vector_push_string.tolist() == [x + ['hi'] for x in FLOAT_VECTORS]
    assert out.json_list_push_string.tolist() == [x + ['hi'] for x in LISTS]
    assert out.json_obj_push_string.tolist() == [None, None, None]

    assert out.text_vector_push_json.tolist(
    ) == [x + [{'x': True}] for x in FLOAT_VECTORS]
    assert out.text_list_push_json.tolist() == [x + [{'x': True}] for x in LISTS]
    assert out.text_obj_push_json.tolist() == [None, None, None]

    assert out.json_vector_push_json.tolist(
    ) == [x + [{'x': True}] for x in FLOAT_VECTORS]
    assert out.json_list_push_json.tolist() == [x + [{'x': True}] for x in LISTS]
    assert out.json_obj_push_json.tolist() == [None, None, None]

    assert out.json_vector_push.tolist() == [x + [{'x': True}] for x in FLOAT_VECTORS]
    assert out.json_list_push.tolist() == [x + [{'x': True}] for x in LISTS]
    assert out.json_obj_push.tolist() == [None, None, None]


def test_json_delete_key(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_vector_delete_key=lambda x: x.text_vector.json_delete_key(1),
        json_vector_delete_key=lambda x: x.json_vector.json_delete_key(1),
        text_list_delete_key=lambda x: x.text_list.json_delete_key(1),
        json_list_delete_key=lambda x: x.json_list.json_delete_key(1),
        text_obj_delete_key=lambda x: x.text_obj.json_delete_key('b'),
        json_obj_delete_key=lambda x: x.json_obj.json_delete_key('b'),

        text_obj_delete_nested_key=lambda x: x.text_obj.json_delete_key('c', 'e'),
        json_obj_delete_nested_key=lambda x: x.json_obj.json_delete_key('c', 'e'),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    del_vector_key = [
        [0.1, -6.7],
        [10.3, 7.8],
        [-2.3, 7.9],
    ]
    del_list_key = [
        [0.1, True],
        [10.3, False],
        [0.3, False],
    ]
    del_obj_key = [
        {'a': 1, 'c': {'d': 100}},
        {'a': 1, 'c': {'d': 105, 'e': True}},
        {'a': 10, 'c': {'d': 111, 'e': 1.234}},
    ]
    del_nested_obj_key = [
        {'a': 1, 'b': 2, 'c': {'d': 100}},
        {'a': 1, 'b': 12, 'c': {'d': 105}},
        {'a': 10, 'b': 25, 'c': {'d': 111}},
    ]

    assert out.text_vector_delete_key.tolist() == del_vector_key
    assert out.json_vector_delete_key.tolist() == del_vector_key
    assert out.text_list_delete_key.tolist() == del_list_key
    assert out.json_list_delete_key.tolist() == del_list_key
    assert out.text_obj_delete_key.tolist() == del_obj_key
    assert out.json_obj_delete_key.tolist() == del_obj_key
    assert out.text_obj_delete_nested_key.tolist() == del_nested_obj_key
    assert out.json_obj_delete_nested_key.tolist() == del_nested_obj_key


def test_json_extract_double(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_vector_extract_double=lambda x: x.text_vector.json_extract_double(1),
        text_list_extract_double=lambda x: x.text_list.json_extract_double(0),
        text_obj_extract_double=lambda x: x.text_obj.json_extract_double('c', 'd'),

        json_vector_extract_double=lambda x: x.json_vector.json_extract_double(1),
        json_list_extract_double=lambda x: x.json_list.json_extract_double(0),
        json_obj_extract_double=lambda x: x.json_obj.json_extract_double('c', 'd'),

        text_vector_extract_string=lambda x: x.text_vector.json_extract_string(1),
        text_list_extract_string=lambda x: x.text_list.json_extract_string(1),
        text_obj_extract_string=lambda x: x.text_obj.json_extract_string('b'),

        json_vector_extract_string=lambda x: x.json_vector.json_extract_string(1),
        json_list_extract_string=lambda x: x.json_list.json_extract_string(1),
        json_obj_extract_string=lambda x: x.json_obj.json_extract_string('b'),

        text_vector_extract_json=lambda x: x.text_vector.json_extract_json(1),
        text_list_extract_json=lambda x: x.text_list.json_extract_json(0),
        text_obj_extract_json=lambda x: x.text_obj.json_extract_json('c'),

        json_vector_extract_json=lambda x: x.json_vector.json_extract_json(1),
        json_list_extract_json=lambda x: x.json_list.json_extract_json(0),
        json_obj_extract_json=lambda x: x.json_obj.json_extract_json('c'),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_vector_extract_double.tolist() == [2.5, -3.3, 6.6]
    assert out.text_list_extract_double.tolist() == [0.1, 10.3, 0.3]
    assert out.text_obj_extract_double.tolist() == [100, 105, 111]

    assert out.json_vector_extract_double.tolist() == [2.5, -3.3, 6.6]
    assert out.json_list_extract_double.tolist() == [0.1, 10.3, 0.3]
    assert out.json_obj_extract_double.tolist() == [100, 105, 111]

    assert out.text_vector_extract_string.tolist() == ['2.5', '-3.3', '6.6']
    assert out.text_list_extract_string.tolist() == ['foo', 'bar', 'baz']
    assert out.text_obj_extract_string.tolist() == ['2', '12', '25']

    assert out.json_vector_extract_string.tolist() == ['2.5', '-3.3', '6.6']
    assert out.json_list_extract_string.tolist() == ['foo', 'bar', 'baz']
    assert out.json_obj_extract_string.tolist() == ['2', '12', '25']

    objs = [
        {'d': 100},
        {'d': 105, 'e': True},
        {'d': 111, 'e': 1.234},
    ]

    assert out.text_vector_extract_json.tolist() == [2.5, -3.3, 6.6]
    assert out.text_list_extract_json.tolist() == [0.1, 10.3, 0.3]
    assert out.text_obj_extract_json.tolist() == objs

    assert out.json_vector_extract_json.tolist() == [2.5, -3.3, 6.6]
    assert out.json_list_extract_json.tolist() == [0.1, 10.3, 0.3]
    assert out.json_obj_extract_json.tolist() == objs


def test_json_get_type(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_get_type_double=lambda x: x.text_vector.json_extract_json(1).json_get_type(),
        json_get_type_double=lambda x: x.json_vector.json_extract_json(1).json_get_type(),

        text_get_type_string=lambda x: x.text_list.json_extract_json(1).json_get_type(),
        json_get_type_string=lambda x: x.json_list.json_extract_json(1).json_get_type(),

        text_get_type_object=lambda x: x.text_obj.json_extract_json('c').json_get_type(),
        json_get_type_object=lambda x: x.json_obj.json_extract_json('c').json_get_type(),

    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_get_type_double.tolist() == ['double', 'double', 'double']
    assert out.json_get_type_double.tolist() == ['double', 'double', 'double']

    assert out.text_get_type_string.tolist() == ['string', 'string', 'string']
    assert out.json_get_type_string.tolist() == ['string', 'string', 'string']

    assert out.text_get_type_object.tolist() == ['object', 'object', 'object']
    assert out.json_get_type_object.tolist() == ['object', 'object', 'object']


def test_json_mask(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_include_mask=lambda x: x.text_obj.json_include_mask('{"b":1}'),
        json_include_mask=lambda x: x.json_obj.json_include_mask('{"b":1}'),

        text_exclude_mask=lambda x: x.text_obj.json_exclude_mask('{"c":1}'),
        json_exclude_mask=lambda x: x.json_obj.json_exclude_mask('{"c":1}'),

        text_include_mask_py=lambda x: x.text_obj.json_include_mask({'b': 1}),
        json_include_mask_py=lambda x: x.json_obj.json_include_mask({'b': 1}),

        text_exclude_mask_py=lambda x: x.text_obj.json_exclude_mask({'c': 1}),
        json_exclude_mask_py=lambda x: x.json_obj.json_exclude_mask({'c': 1}),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_include_mask.tolist() == [{'b': 2}, {'b': 12}, {'b': 25}]
    assert out.json_include_mask.tolist() == [{'b': 2}, {'b': 12}, {'b': 25}]

    assert out.text_exclude_mask.tolist() == [
        {'a': 1, 'b': 2},
        {'a': 1, 'b': 12},
        {'a': 10, 'b': 25},
    ]
    assert out.json_exclude_mask.tolist() == [
        {'a': 1, 'b': 2},
        {'a': 1, 'b': 12},
        {'a': 10, 'b': 25},
    ]

    assert out.text_include_mask_py.tolist() == [{'b': 2}, {'b': 12}, {'b': 25}]
    assert out.json_include_mask_py.tolist() == [{'b': 2}, {'b': 12}, {'b': 25}]

    assert out.text_exclude_mask_py.tolist() == [
        {'a': 1, 'b': 2},
        {'a': 1, 'b': 12},
        {'a': 10, 'b': 25},
    ]
    assert out.json_exclude_mask_py.tolist() == [
        {'a': 1, 'b': 2},
        {'a': 1, 'b': 12},
        {'a': 10, 'b': 25},
    ]


def test_json_keys(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_keys=lambda x: x.text_obj.json_keys(),
        json_keys=lambda x: x.json_obj.json_keys(),

        text_nested_keys=lambda x: x.text_obj.json_keys('c'),
        json_nested_keys=lambda x: x.json_obj.json_keys('c'),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_keys.tolist() == [['a', 'b', 'c'], ['a', 'b', 'c'], ['a', 'b', 'c']]
    assert out.json_keys.tolist() == [['a', 'b', 'c'], ['a', 'b', 'c'], ['a', 'b', 'c']]

    assert out.text_nested_keys.tolist() == [['d'], ['d', 'e'], ['d', 'e']]
    assert out.json_nested_keys.tolist() == [['d'], ['d', 'e'], ['d', 'e']]


def test_json_length(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_length=lambda x: x.text_obj.json_length(),
        json_length=lambda x: x.json_obj.json_length(),

        text_nested_length=lambda x: x.text_obj.json_extract_json('c').json_length(),
        json_nested_length=lambda x: x.json_obj.json_extract_json('c').json_length(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_length.tolist() == [3, 3, 3]
    assert out.json_length.tolist() == [3, 3, 3]

    assert out.text_nested_length.tolist() == [1, 2, 2]
    assert out.json_nested_length.tolist() == [1, 2, 2]


def test_json_pretty(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_pretty=lambda x: x.text_obj.json_pretty(),
        json_pretty=lambda x: x.json_obj.json_pretty(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    pretty = [
        '''{
  "a": 1,
  "b": 2,
  "c": {
    "d": 100
  }
}''',
        '''{
  "a": 1,
  "b": 12,
  "c": {
    "d": 105,
    "e": true
  }
}''',
        '''{
  "a": 10,
  "b": 25,
  "c": {
    "d": 111,
    "e": 1.234
  }
}''',
    ]

    assert out.text_pretty.tolist() == pretty
    assert out.json_pretty.tolist() == pretty


def test_json_set_double(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_set_double=lambda x: x.text_obj.json_set_double('f', 3.14),
        json_set_double=lambda x: x.json_obj.json_set_double('f', 3.14),

        text_set_array_double=lambda x: x.text_list.json_set_double(1, 3.14),
        json_set_array_double=lambda x: x.json_list.json_set_double(1, 3.14),

        text_set_nested_double=lambda x: x.text_obj.json_set_double('c', 'f', 3.14),
        json_set_nested_double=lambda x: x.json_obj.json_set_double('c', 'f', 3.14),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    values = out.text_set_double.tolist()
    assert values[0]['f'] == 3.14
    assert values[1]['f'] == 3.14
    assert values[2]['f'] == 3.14

    values = out.json_set_double.tolist()
    assert values[0]['f'] == 3.14
    assert values[1]['f'] == 3.14
    assert values[2]['f'] == 3.14

    values = out.text_set_nested_double.tolist()
    assert values[0]['c']['f'] == 3.14
    assert values[1]['c']['f'] == 3.14
    assert values[2]['c']['f'] == 3.14

    values = out.json_set_nested_double.tolist()
    assert values[0]['c']['f'] == 3.14
    assert values[1]['c']['f'] == 3.14
    assert values[2]['c']['f'] == 3.14

    assert out.text_set_array_double.tolist() == [
        [0.1, 3.14, True],
        [10.3, 3.14, False],
        [0.3, 3.14, False],
    ]
    assert out.json_set_array_double.tolist() == [
        [0.1, 3.14, True],
        [10.3, 3.14, False],
        [0.3, 3.14, False],
    ]


def test_json_set_string(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_set_string=lambda x: x.text_obj.json_set_string('f', '3.14'),
        json_set_string=lambda x: x.json_obj.json_set_string('f', '3.14'),

        text_set_array_string=lambda x: x.text_list.json_set_string(1, '3.14'),
        json_set_array_string=lambda x: x.json_list.json_set_string(1, '3.14'),

        text_set_nested_string=lambda x: x.text_obj.json_set_string('c', 'f', '3.14'),
        json_set_nested_string=lambda x: x.json_obj.json_set_string('c', 'f', '3.14'),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    values = out.text_set_string.tolist()
    assert values[0]['f'] == '3.14'
    assert values[1]['f'] == '3.14'
    assert values[2]['f'] == '3.14'

    values = out.json_set_string.tolist()
    assert values[0]['f'] == '3.14'
    assert values[1]['f'] == '3.14'
    assert values[2]['f'] == '3.14'

    values = out.text_set_nested_string.tolist()
    assert values[0]['c']['f'] == '3.14'
    assert values[1]['c']['f'] == '3.14'
    assert values[2]['c']['f'] == '3.14'

    values = out.json_set_nested_string.tolist()
    assert values[0]['c']['f'] == '3.14'
    assert values[1]['c']['f'] == '3.14'
    assert values[2]['c']['f'] == '3.14'

    assert out.text_set_array_string.tolist() == [
        [0.1, '3.14', True],
        [10.3, '3.14', False],
        [0.3, '3.14', False],
    ]
    assert out.json_set_array_string.tolist() == [
        [0.1, '3.14', True],
        [10.3, '3.14', False],
        [0.3, '3.14', False],
    ]


def test_json_set_json(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_set_json=lambda x: x.text_obj.json_set_json('f', 'true'),
        json_set_json=lambda x: x.json_obj.json_set_json('f', 'true'),

        text_set_json_obj=lambda x: x.text_obj.json_set_json('f', '{"x":1000}'),
        json_set_json_obj=lambda x: x.json_obj.json_set_json('f', '{"x":1000}'),

        text_set_array_json=lambda x: x.text_list.json_set_json(1, 'true'),
        json_set_array_json=lambda x: x.json_list.json_set_json(1, 'true'),

        text_set_nested_json=lambda x: x.text_obj.json_set_json('c', 'f', 'true'),
        json_set_nested_json=lambda x: x.json_obj.json_set_json('c', 'f', 'true'),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    values = out.text_set_json.tolist()
    assert values[0]['f'] is True
    assert values[1]['f'] is True
    assert values[2]['f'] is True

    values = out.json_set_json.tolist()
    assert values[0]['f'] is True
    assert values[1]['f'] is True
    assert values[2]['f'] is True

    values = out.text_set_json_obj.tolist()
    assert values[0]['f'] == dict(x=1000)
    assert values[1]['f'] == dict(x=1000)
    assert values[2]['f'] == dict(x=1000)

    values = out.json_set_json_obj.tolist()
    assert values[0]['f'] == dict(x=1000)
    assert values[1]['f'] == dict(x=1000)
    assert values[2]['f'] == dict(x=1000)

    values = out.text_set_nested_json.tolist()
    assert values[0]['c']['f'] is True
    assert values[1]['c']['f'] is True
    assert values[2]['c']['f'] is True

    values = out.json_set_nested_json.tolist()
    assert values[0]['c']['f'] is True
    assert values[1]['c']['f'] is True
    assert values[2]['c']['f'] is True

    assert out.text_set_array_json.tolist() == [
        [0.1, True, True],
        [10.3, True, False],
        [0.3, True, False],
    ]
    assert out.json_set_array_json.tolist() == [
        [0.1, True, True],
        [10.3, True, False],
        [0.3, True, False],
    ]


def test_json_set(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_set=lambda x: x.text_obj.json_set('f', True),
        json_set=lambda x: x.json_obj.json_set('f', True),

        text_set_obj=lambda x: x.text_obj.json_set('f', {'x': 1000}),
        json_set_obj=lambda x: x.json_obj.json_set('f', {'x': 1000}),

        text_set_array=lambda x: x.text_list.json_set(1, True),
        json_set_array=lambda x: x.json_list.json_set(1, True),

        text_set_nested=lambda x: x.text_obj.json_set('c', 'f', True),
        json_set_nested=lambda x: x.json_obj.json_set('c', 'f', True),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    values = out.text_set.tolist()
    assert values[0]['f'] is True
    assert values[1]['f'] is True
    assert values[2]['f'] is True

    values = out.json_set.tolist()
    assert values[0]['f'] is True
    assert values[1]['f'] is True
    assert values[2]['f'] is True

    values = out.text_set_obj.tolist()
    assert values[0]['f'] == dict(x=1000)
    assert values[1]['f'] == dict(x=1000)
    assert values[2]['f'] == dict(x=1000)

    values = out.json_set_obj.tolist()
    assert values[0]['f'] == dict(x=1000)
    assert values[1]['f'] == dict(x=1000)
    assert values[2]['f'] == dict(x=1000)

    values = out.text_set_nested.tolist()
    assert values[0]['c']['f'] is True
    assert values[1]['c']['f'] is True
    assert values[2]['c']['f'] is True

    values = out.json_set_nested.tolist()
    assert values[0]['c']['f'] is True
    assert values[1]['c']['f'] is True
    assert values[2]['c']['f'] is True

    assert out.text_set_array.tolist() == [
        [0.1, True, True],
        [10.3, True, False],
        [0.3, True, False],
    ]
    assert out.json_set_array.tolist() == [
        [0.1, True, True],
        [10.3, True, False],
        [0.3, True, False],
    ]


def test_json_splice(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        text_splice_double=lambda x: x.text_list.json_splice_double(1, 3, 3.14),
        json_splice_double=lambda x: x.json_list.json_splice_double(1, 3, 3.14),

        text_splice_string=lambda x: x.text_list.json_splice_string(1, 3, '3.14'),
        json_splice_string=lambda x: x.json_list.json_splice_string(1, 3, '3.14'),

        text_splice_json=lambda x: x.text_list.json_splice_json(1, 3, '{"x": 3.14}'),
        json_splice_json=lambda x: x.json_list.json_splice_json(1, 3, '{"x": 3.14}'),

        text_splice=lambda x: x.text_list.json_splice(1, 3, {'x': 3.14}),
        json_splice=lambda x: x.json_list.json_splice(1, 3, {'x': 3.14}),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.text_splice_double.tolist() == [[0.1, 3.14], [10.3, 3.14], [0.3, 3.14]]
    assert out.json_splice_double.tolist() == [[0.1, 3.14], [10.3, 3.14], [0.3, 3.14]]

    assert out.text_splice_string.tolist() == [
        [0.1, '3.14'], [
            10.3, '3.14',
        ], [0.3, '3.14'],
    ]
    assert out.json_splice_string.tolist() == [
        [0.1, '3.14'], [
            10.3, '3.14',
        ], [0.3, '3.14'],
    ]

    assert out.text_splice_json.tolist() == [
        [0.1, {'x': 3.14}],
        [10.3, {'x': 3.14}],
        [0.3, {'x': 3.14}],
    ]
    assert out.json_splice_json.tolist() == [
        [0.1, {'x': 3.14}],
        [10.3, {'x': 3.14}],
        [0.3, {'x': 3.14}],
    ]

    assert out.text_splice.tolist() == [
        [0.1, {'x': 3.14}],
        [10.3, {'x': 3.14}],
        [0.3, {'x': 3.14}],
    ]
    assert out.json_splice.tolist() == [
        [0.1, {'x': 3.14}],
        [10.3, {'x': 3.14}],
        [0.3, {'x': 3.14}],
    ]


def test_bitcount(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        int_bitcount=lambda x: x.int_c.bit_count(),
        bigint_bitcount=lambda x: x.bigint_c.bit_count(),
        float_bitcount=lambda x: x.float_c.bit_count(),
        double_bitcount=lambda x: x.double_c.bit_count(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    out['py_int_bitcount'] = out['int_c'].apply(lambda x: format(x, 'b').count('1'))
    out['py_bigint_bitcount'] = out['bigint_c'].apply(lambda x: format(x, 'b').count('1'))
    out['py_float_bitcount'] = out['float_c'].apply(
        lambda x: format(round(x), 'b').count('1'),
    )
    out['py_double_bitcount'] = out['double_c'].apply(
        lambda x: format(round(x), 'b').count('1'),
    )

    assert out.int_bitcount.tolist() == out.py_int_bitcount.tolist()
    assert out.bigint_bitcount.tolist() == out.py_bigint_bitcount.tolist()
    assert out.float_bitcount.tolist() == out.py_float_bitcount.tolist()
    assert out.double_bitcount.tolist() == out.py_double_bitcount.tolist()


def test_conv(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        int_hex=lambda x: x.int_c.conv(10, 16),
        bigint_hex=lambda x: x.bigint_c.conv(10, 16),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    out['py_int_hex'] = out['int_c'].apply(lambda x: np.base_repr(x, base=16))
    out['py_bigint_hex'] = out['bigint_c'].apply(lambda x: np.base_repr(x, base=16))

    assert out.int_hex.tolist() == out.py_int_hex.tolist()
    assert out.bigint_hex.tolist() == out.py_bigint_hex.tolist()


def test_sigmoid(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        int_sigmoid=lambda x: x.int_c.sigmoid(),
        bigint_sigmoid=lambda x: x.bigint_c.sigmoid(),
        float_sigmoid=lambda x: x.float_c.sigmoid(),
        double_sigmoid=lambda x: x.double_c.sigmoid(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    def sigmoid(x: Any) -> Any:
        return np.where(x < 0, np.exp(x)/(1 + np.exp(x)), 1/(1 + np.exp(-x)))

    out['py_int_sigmoid'] = out['int_c'].apply(sigmoid)
    out['py_bigint_sigmoid'] = out['bigint_c'].apply(sigmoid)
    out['py_float_sigmoid'] = out['float_c'].apply(sigmoid)
    out['py_double_sigmoid'] = out['double_c'].apply(sigmoid)

    assert out.int_sigmoid.tolist() == out.py_int_sigmoid.tolist()
    assert out.bigint_sigmoid.tolist() == out.py_bigint_sigmoid.tolist()
    assert out.float_sigmoid.tolist() == out.py_float_sigmoid.tolist()
    assert out.double_sigmoid.tolist() == out.py_double_sigmoid.tolist()


def test_to_number(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        parse_int=lambda x: x.int_c.cast('string').to_number(),
        parse_bigint=lambda x: x.bigint_c.cast('string').to_number(),
        parse_float=lambda x: x.float_c.cast('string').to_number(),
        parse_double=lambda x: x.double_c.cast('string').to_number(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    tm.assert_series_equal(
        out.parse_int, pd.Series(
            [41.0, 44.0, 30.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.parse_bigint, pd.Series(
            [12.0, 46.0, 65.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.parse_float, pd.Series(
            [88.8983, 11.4726, 13.5833],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.parse_double, pd.Series(
            [49.002318, 87.812455, 58.052145],
        ), check_names=False,
    )


def test_trunc(con: Any) -> None:
    tbl = con.table('datatypes')

    out = tbl.mutate(
        trunc_float=lambda x: x.float_c.trunc(),
        trunc_double=lambda x: x.double_c.trunc(),

        trunc_float_2=lambda x: x.float_c.trunc(2),
        trunc_double_2=lambda x: x.double_c.trunc(2),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    tm.assert_series_equal(
        out.trunc_float, pd.Series(
            [88.0, 11.0, 13.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.trunc_double, pd.Series(
            [49.0, 87.0, 58.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.trunc_float_2, pd.Series(
            [88.89, 11.47, 13.58],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.trunc_double_2, pd.Series(
            [49.00, 87.81, 58.05],
        ), check_names=False,
    )


def test_truncate(con: Any) -> None:
    tbl = con.table('datatypes')

    with pytest.raises(TypeError):
        tbl.float_c.truncate()

    out = tbl.mutate(
        truncate_float_2=lambda x: x.float_c.truncate(2),
        truncate_double_2=lambda x: x.double_c.truncate(2),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    tm.assert_series_equal(
        out.truncate_float_2, pd.Series(
            [88.89, 11.47, 13.58],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.truncate_double_2, pd.Series(
            [49.00, 87.81, 58.05],
        ), check_names=False,
    )


def test_dot_product(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        dp=lambda x: x.vec1.dot_product(x.vec2),
        dp_f32=lambda x: x.vec1_f32.dot_product_f32(x.vec2_f32),
        dp_f64=lambda x: x.vec1_f64.dot_product_f64(x.vec2_f64),
        dp_i8=lambda x: x.vec1_i8.dot_product_i8(x.vec2_i8),
        dp_i16=lambda x: x.vec1_i16.dot_product_i16(x.vec2_i16),
        dp_i32=lambda x: x.vec1_i32.dot_product_i32(x.vec2_i32),
        dp_i64=lambda x: x.vec1_i64.dot_product_i64(x.vec2_i64),

        dp_lit=lambda x: x.vec1.dot_product([1, 2, 3]),
        dp_f32_lit=lambda x: x.vec1_f32.dot_product_f32([1, 2, 3]),
        dp_f64_lit=lambda x: x.vec1_f64.dot_product_f64([1, 2, 3]),
        dp_i8_lit=lambda x: x.vec1_i8.dot_product_i8([1, 2, 3]),
        dp_i16_lit=lambda x: x.vec1_i16.dot_product_i16([1, 2, 3]),
        dp_i32_lit=lambda x: x.vec1_i32.dot_product_i32([1, 2, 3]),
        dp_i64_lit=lambda x: x.vec1_i64.dot_product_i64([1, 2, 3]),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    tm.assert_series_equal(
        out.dp, pd.Series(
            [51.149994, 177.820007, 111.259995],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_f32, pd.Series(
            [51.149994, 177.820007, 111.259995],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_f64, pd.Series(
            [51.15, 177.82, 111.26],
        ), check_names=False,
    )
    assert out.dp_i8.tolist() == [58.0, 173.0, 117.0]
    assert out.dp_i16.tolist() == [58.0, 173.0, 117.0]
    assert out.dp_i32.tolist() == [58.0, 173.0, 117.0]
    assert out.dp_i64.tolist() == [58.0, 173.0, 117.0]

    tm.assert_series_equal(
        out.dp_lit, pd.Series(
            [-14.999998, 27.100002, 34.599998],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_f32_lit, pd.Series(
            [-14.999998, 27.100002, 34.599998],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_f64_lit, pd.Series(
            [-15.0, 27.1, 34.6],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_i8_lit, pd.Series(
            [-15.0, 28.0, 36.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_i16_lit, pd.Series(
            [-15.0, 28.0, 36.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_i32_lit, pd.Series(
            [-15.0, 28.0, 36.0],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.dp_i64_lit, pd.Series(
            [-15.0, 28.0, 36.0],
        ), check_names=False,
    )


def test_euclidean_distance(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        ed=lambda x: x.vec1.euclidean_distance(x.vec2),
        ed_f32=lambda x: x.vec1_f32.euclidean_distance_f32(x.vec2_f32),
        ed_f64=lambda x: x.vec1_f64.euclidean_distance_f64(x.vec2_f64),
        ed_i8=lambda x: x.vec1_i8.euclidean_distance_i8(x.vec2_i8),
        ed_i16=lambda x: x.vec1_i16.euclidean_distance_i16(x.vec2_i16),
        ed_i32=lambda x: x.vec1_i32.euclidean_distance_i32(x.vec2_i32),
        ed_i64=lambda x: x.vec1_i64.euclidean_distance_i64(x.vec2_i64),

        ed_lit=lambda x: x.vec1.euclidean_distance([1, 2, 3]),
        ed_f32_lit=lambda x: x.vec1_f32.euclidean_distance_f32([1, 2, 3]),
        ed_f64_lit=lambda x: x.vec1_f64.euclidean_distance_f64([1, 2, 3]),
        ed_i8_lit=lambda x: x.vec1_i8.euclidean_distance_i8([1, 2, 3]),
        ed_i16_lit=lambda x: x.vec1_i16.euclidean_distance_i16([1, 2, 3]),
        ed_i32_lit=lambda x: x.vec1_i32.euclidean_distance_i32([1, 2, 3]),
        ed_i64_lit=lambda x: x.vec1_i64.euclidean_distance_i64([1, 2, 3]),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    tm.assert_series_equal(out.ed, pd.Series([0.0, 0.0, 0.0]), check_names=False)
    tm.assert_series_equal(out.ed_f32, pd.Series([0.0, 0.0, 0.0]), check_names=False)
    tm.assert_series_equal(out.ed_f64, pd.Series([0.0, 0.0, 0.0]), check_names=False)
    assert out.ed_i8.tolist() == [0.0, 0.0, 0.0]
    assert out.ed_i16.tolist() == [0.0, 0.0, 0.0]
    assert out.ed_i32.tolist() == [0.0, 0.0, 0.0]
    assert out.ed_i64.tolist() == [0.0, 0.0, 0.0]

    tm.assert_series_equal(
        out.ed_lit, pd.Series(
            [9.754486, 11.731156, 7.487322],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_f32_lit, pd.Series(
            [9.754486, 11.731156, 7.487322],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_f64_lit, pd.Series(
            [9.754486, 11.731156, 7.487322],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_i8_lit, pd.Series(
            [10.099505, 11.445523, 7.681146],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_i16_lit, pd.Series(
            [10.099505, 11.445523, 7.681146],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_i32_lit, pd.Series(
            [10.099505, 11.445523, 7.681146],
        ), check_names=False,
    )
    tm.assert_series_equal(
        out.ed_i64_lit, pd.Series(
            [10.099505, 11.445523, 7.681146],
        ), check_names=False,
    )


def test_scalar_vector_mul(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        ed=lambda x: x.vec1.scalar_vector_mul(3).json_array_unpack(),
        ed_f32=lambda x: x.vec1_f32.scalar_vector_mul_f32(3).json_array_unpack_f32(),
        ed_f64=lambda x: x.vec1_f64.scalar_vector_mul_f64(3).json_array_unpack_f64(),
        ed_i8=lambda x: x.vec1_i8.scalar_vector_mul_i8(3).json_array_unpack_i8(),
        ed_i16=lambda x: x.vec1_i16.scalar_vector_mul_i16(3).json_array_unpack_i16(),
        ed_i32=lambda x: x.vec1_i32.scalar_vector_mul_i32(3).json_array_unpack_i32(),
        ed_i64=lambda x: x.vec1_i64.scalar_vector_mul_i64(3).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([[y * 3 for y in x] for x in FLOAT_VECTORS])
    ints = pd.Series([[y * 3 for y in x] for x in INT_VECTORS])

    tm.assert_series_equal(out.ed, floats, check_names=False)
    tm.assert_series_equal(out.ed_f32, floats, check_names=False)
    tm.assert_series_equal(out.ed_f64, floats, check_names=False)
    tm.assert_series_equal(out.ed_i8, ints, check_names=False)
    tm.assert_series_equal(out.ed_i16, ints, check_names=False)
    tm.assert_series_equal(out.ed_i32, ints, check_names=False)
    tm.assert_series_equal(out.ed_i64, ints, check_names=False)


def test_vector_add(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        va=lambda x: x.vec1.vector_add(x.vec2).json_array_unpack(),
        va_f32=lambda x: x.vec1_f32.vector_add_f32(x.vec2_f32).json_array_unpack_f32(),
        va_f64=lambda x: x.vec1_f64.vector_add_f64(x.vec2_f64).json_array_unpack_f64(),
        va_i8=lambda x: x.vec1_i8.vector_add_i8(x.vec2_i8).json_array_unpack_i8(),
        va_i16=lambda x: x.vec1_i16.vector_add_i16(x.vec2_i16).json_array_unpack_i16(),
        va_i32=lambda x: x.vec1_i32.vector_add_i32(x.vec2_i32).json_array_unpack_i32(),
        va_i64=lambda x: x.vec1_i64.vector_add_i64(x.vec2_i64).json_array_unpack_i64(),

        va_lit=lambda x: x.vec1.vector_add([1.0, 2.0, 3.0]).json_array_unpack(),
        va_f32_lit=lambda x: x.vec1_f32.vector_add_f32(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_f32(),
        va_f64_lit=lambda x: x.vec1_f64.vector_add_f64(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_f64(),
        va_i8_lit=lambda x: x.vec1_i8.vector_add_i8(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i8(),
        va_i16_lit=lambda x: x.vec1_i16.vector_add_i16(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i16(),
        va_i32_lit=lambda x: x.vec1_i32.vector_add_i32(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i32(),
        va_i64_lit=lambda x: x.vec1_i64.vector_add_i64(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([
        [0.200000003, 5, -13.3999996],
        [20.6000004, -6.5999999, 15.6000004],
        [-4.5999999, 13.1999998, 15.8000002],
    ])
    ints = pd.Series([[0, 6, -14], [20, -6, 16], [-4, 14, 16]])

    tm.assert_series_equal(out.va, floats, check_names=False)
    tm.assert_series_equal(out.va_f32, floats, check_names=False)
    tm.assert_series_equal(out.va_f64, floats, check_names=False)
    tm.assert_series_equal(out.va_i8, ints, check_names=False)
    tm.assert_series_equal(out.va_i16, ints, check_names=False)
    tm.assert_series_equal(out.va_i32, ints, check_names=False)
    tm.assert_series_equal(out.va_i64, ints, check_names=False)

    floats = pd.Series([[x[0] + 1, x[1] + 2, x[2] + 3] for x in FLOAT_VECTORS])
    ints = pd.Series([[x[0] + 1, x[1] + 2, x[2] + 3] for x in INT_VECTORS])

    tm.assert_series_equal(out.va_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_f32_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_f64_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_i8_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i16_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i32_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i64_lit, ints, check_names=False)


def test_vector_elements_sum(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        vs=lambda x: x.vec1.vector_elements_sum(),
        vs_f32=lambda x: x.vec1_f32.vector_elements_sum_f32(),
        vs_f64=lambda x: x.vec1_f64.vector_elements_sum_f64(),
        vs_i8=lambda x: x.vec1_i8.vector_elements_sum_i8(),
        vs_i16=lambda x: x.vec1_i16.vector_elements_sum_i16(),
        vs_i32=lambda x: x.vec1_i32.vector_elements_sum_i32(),
        vs_i64=lambda x: x.vec1_i64.vector_elements_sum_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([-4.1, 14.8, 12.2])
    ints = pd.Series([-4.0, 15.0, 13.0])

    tm.assert_series_equal(out.vs, floats, check_names=False)
    tm.assert_series_equal(out.vs_f32, floats, check_names=False)
    tm.assert_series_equal(out.vs_f64, floats, check_names=False)
    tm.assert_series_equal(out.vs_i8, ints, check_names=False)
    tm.assert_series_equal(out.vs_i16, ints, check_names=False)
    tm.assert_series_equal(out.vs_i32, ints, check_names=False)
    tm.assert_series_equal(out.vs_i64, ints, check_names=False)


def test_vector_kth_element(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        ve=lambda x: x.vec1.vector_kth_element(1),
        ve_f32=lambda x: x.vec1_f32.vector_kth_element_f32(1),
        ve_f64=lambda x: x.vec1_f64.vector_kth_element_f64(1),
        ve_i8=lambda x: x.vec1_i8.vector_kth_element_i8(1),
        ve_i16=lambda x: x.vec1_i16.vector_kth_element_i16(1),
        ve_i32=lambda x: x.vec1_i32.vector_kth_element_i32(1),
        ve_i64=lambda x: x.vec1_i64.vector_kth_element_i64(1),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([2.5, -3.3, 6.6])
    ints = pd.Series([3.0, -3.0, 7.0])

    tm.assert_series_equal(out.ve, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_f32, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_f64, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_i8, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_i16, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_i32, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.ve_i64, ints, check_names=False, check_dtype=False)


def test_vector_mul(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        vm=lambda x: x.vec1.vector_mul(x.vec2).json_array_unpack(),
        vm_f32=lambda x: x.vec1_f32.vector_mul_f32(x.vec2_f32).json_array_unpack_f32(),
        vm_f64=lambda x: x.vec1_f64.vector_mul_f64(x.vec2_f64).json_array_unpack_f64(),
        vm_i8=lambda x: x.vec1_i8.vector_mul_i8(x.vec2_i8).json_array_unpack_i8(),
        vm_i16=lambda x: x.vec1_i16.vector_mul_i16(x.vec2_i16).json_array_unpack_i16(),
        vm_i32=lambda x: x.vec1_i32.vector_mul_i32(x.vec2_i32).json_array_unpack_i32(),
        vm_i64=lambda x: x.vec1_i64.vector_mul_i64(x.vec2_i64).json_array_unpack_i64(),

        vm_lit=lambda x: x.vec1.vector_mul([1, 2, 3]).json_array_unpack(),
        vm_f32_lit=lambda x: x.vec1_f32.vector_mul_f32([1, 2, 3]).json_array_unpack_f32(),
        vm_f64_lit=lambda x: x.vec1_f64.vector_mul_f64([1, 2, 3]).json_array_unpack_f64(),
        vm_i8_lit=lambda x: x.vec1_i8.vector_mul_i8([1, 2, 3]).json_array_unpack_i8(),
        vm_i16_lit=lambda x: x.vec1_i16.vector_mul_i16([1, 2, 3]).json_array_unpack_i16(),
        vm_i32_lit=lambda x: x.vec1_i32.vector_mul_i32([1, 2, 3]).json_array_unpack_i32(),
        vm_i64_lit=lambda x: x.vec1_i64.vector_mul_i64([1, 2, 3]).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([
        [0.0100000007, 6.25, 44.8899956],
        [106.090004, 10.8899994, 60.840004],
        [5.28999996, 43.5599976, 62.4099998],
    ])
    ints = pd.Series([[0, 9, 49], [100, 9, 64], [4, 49, 64]])

    # TODO: Bug in vector_mul?!?
#   tm.assert_series_equal(out.vm, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_f32, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_f64, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i8, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i16, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i32, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i64, ints, check_names=False, check_dtype=False)

    floats = pd.Series([
        [0.100000001, 5, -20.0999985],
        [10.3000002, -6.5999999, 23.4000015],
        [-2.29999995, 13.1999998, 23.7000008],
    ])
    ints = pd.Series([[0, 6, -21], [10, -6, 24], [-2, 14, 24]])

    # TODO: Bug in vector_mul?!?
#   tm.assert_series_equal(out.vm_lit, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_f32_lit, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_f64_lit, floats, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i8_lit, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i16_lit, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i32_lit, ints, check_names=False, check_dtype=False)
    tm.assert_series_equal(out.vm_i64_lit, ints, check_names=False, check_dtype=False)


def test_vector_num_elements(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        vm=lambda x: x.vec1.vector_num_elements(),
        vm_f32=lambda x: x.vec1_f32.vector_num_elements_f32(),
        vm_f64=lambda x: x.vec1_f64.vector_num_elements_f64(),
        vm_i8=lambda x: x.vec1_i8.vector_num_elements_i8(),
        vm_i16=lambda x: x.vec1_i16.vector_num_elements_i16(),
        vm_i32=lambda x: x.vec1_i32.vector_num_elements_i32(),
        vm_i64=lambda x: x.vec1_i64.vector_num_elements_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    assert out.vm.tolist() == [3, 3, 3]
    assert out.vm_f32.tolist() == [3, 3, 3]
    assert out.vm_f64.tolist() == [3, 3, 3]
    assert out.vm_i8.tolist() == [3, 3, 3]
    assert out.vm_i16.tolist() == [3, 3, 3]
    assert out.vm_i32.tolist() == [3, 3, 3]
    assert out.vm_i64.tolist() == [3, 3, 3]


def test_vector_sort(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        vs=lambda x: x.vec1.vector_sort().json_array_unpack(),
        vs_f32=lambda x: x.vec1_f32.vector_sort_f32().json_array_unpack_f32(),
        vs_f64=lambda x: x.vec1_f64.vector_sort_f64().json_array_unpack_f64(),
        vs_i8=lambda x: x.vec1_i8.vector_sort_i8().json_array_unpack_i8(),
        vs_i16=lambda x: x.vec1_i16.vector_sort_i16().json_array_unpack_i16(),
        vs_i32=lambda x: x.vec1_i32.vector_sort_i32().json_array_unpack_i32(),
        vs_i64=lambda x: x.vec1_i64.vector_sort_i64().json_array_unpack_i64(),

        vs_dir=lambda x: x.vec1.vector_sort('desc').json_array_unpack(),
        vs_f32_dir=lambda x: x.vec1_f32.vector_sort_f32('desc').json_array_unpack_f32(),
        vs_f64_dir=lambda x: x.vec1_f64.vector_sort_f64('desc').json_array_unpack_f64(),
        vs_i8_dir=lambda x: x.vec1_i8.vector_sort_i8('desc').json_array_unpack_i8(),
        vs_i16_dir=lambda x: x.vec1_i16.vector_sort_i16('desc').json_array_unpack_i16(),
        vs_i32_dir=lambda x: x.vec1_i32.vector_sort_i32('desc').json_array_unpack_i32(),
        vs_i64_dir=lambda x: x.vec1_i64.vector_sort_i64('desc').json_array_unpack_i64(),

        vs_idir=lambda x: x.vec1.vector_sort(ibis.desc).json_array_unpack(),
        vs_f32_idir=lambda x: x.vec1_f32.vector_sort_f32(
            ibis.desc,
        ).json_array_unpack_f32(),
        vs_f64_idir=lambda x: x.vec1_f64.vector_sort_f64(
            ibis.desc,
        ).json_array_unpack_f64(),
        vs_i8_idir=lambda x: x.vec1_i8.vector_sort_i8(ibis.desc).json_array_unpack_i8(),
        vs_i16_idir=lambda x: x.vec1_i16.vector_sort_i16(
            ibis.desc,
        ).json_array_unpack_i16(),
        vs_i32_idir=lambda x: x.vec1_i32.vector_sort_i32(
            ibis.desc,
        ).json_array_unpack_i32(),
        vs_i64_idir=lambda x: x.vec1_i64.vector_sort_i64(
            ibis.desc,
        ).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([
        [-6.69999981, 0.100000001, 2.5],
        [-3.29999995, 7.80000019, 10.3000002],
        [-2.29999995, 6.5999999, 7.9000001],
    ])
    ints = pd.Series([[-7, 0, 3], [-3, 8, 10], [-2, 7, 8]])

    tm.assert_series_equal(out.vs, floats, check_names=False)
    tm.assert_series_equal(out.vs_f32, floats, check_names=False)
    tm.assert_series_equal(out.vs_f64, floats, check_names=False)
    tm.assert_series_equal(out.vs_i8, ints, check_names=False)
    tm.assert_series_equal(out.vs_i16, ints, check_names=False)
    tm.assert_series_equal(out.vs_i32, ints, check_names=False)
    tm.assert_series_equal(out.vs_i64, ints, check_names=False)

    floats = pd.Series([
        list(reversed([-6.69999981, 0.100000001, 2.5])),
        list(reversed([-3.29999995, 7.80000019, 10.3000002])),
        list(reversed([-2.29999995, 6.5999999, 7.9000001])),
    ])
    ints = pd.Series([
        list(reversed([-7, 0, 3])),
        list(reversed([-3, 8, 10])),
        list(reversed([-2, 7, 8])),
    ])

    tm.assert_series_equal(out.vs_dir, floats, check_names=False)
    tm.assert_series_equal(out.vs_f32_dir, floats, check_names=False)
    tm.assert_series_equal(out.vs_f64_dir, floats, check_names=False)
    tm.assert_series_equal(out.vs_i8_dir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i16_dir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i32_dir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i64_dir, ints, check_names=False)

    tm.assert_series_equal(out.vs_idir, floats, check_names=False)
    tm.assert_series_equal(out.vs_f32_idir, floats, check_names=False)
    tm.assert_series_equal(out.vs_f64_idir, floats, check_names=False)
    tm.assert_series_equal(out.vs_i8_idir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i16_idir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i32_idir, ints, check_names=False)
    tm.assert_series_equal(out.vs_i64_dir, ints, check_names=False)


def test_vector_sub(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        va=lambda x: x.vec1.vector_sub(x.vec2).json_array_unpack(),
        va_f32=lambda x: x.vec1_f32.vector_sub_f32(x.vec2_f32).json_array_unpack_f32(),
        va_f64=lambda x: x.vec1_f64.vector_sub_f64(x.vec2_f64).json_array_unpack_f64(),
        va_i8=lambda x: x.vec1_i8.vector_sub_i8(x.vec2_i8).json_array_unpack_i8(),
        va_i16=lambda x: x.vec1_i16.vector_sub_i16(x.vec2_i16).json_array_unpack_i16(),
        va_i32=lambda x: x.vec1_i32.vector_sub_i32(x.vec2_i32).json_array_unpack_i32(),
        va_i64=lambda x: x.vec1_i64.vector_sub_i64(x.vec2_i64).json_array_unpack_i64(),

        va_lit=lambda x: x.vec1.vector_sub([1.0, 2.0, 3.0]).json_array_unpack(),
        va_f32_lit=lambda x: x.vec1_f32.vector_sub_f32(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_f32(),
        va_f64_lit=lambda x: x.vec1_f64.vector_sub_f64(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_f64(),
        va_i8_lit=lambda x: x.vec1_i8.vector_sub_i8(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i8(),
        va_i16_lit=lambda x: x.vec1_i16.vector_sub_i16(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i16(),
        va_i32_lit=lambda x: x.vec1_i32.vector_sub_i32(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i32(),
        va_i64_lit=lambda x: x.vec1_i64.vector_sub_i64(
            [1.0, 2.0, 3.0],
        ).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0],
    ])
    ints = pd.Series([[0, 0, 0], [0, 0, 0], [0, 0, 0]])

    tm.assert_series_equal(out.va, floats, check_names=False)
    tm.assert_series_equal(out.va_f32, floats, check_names=False)
    tm.assert_series_equal(out.va_f64, floats, check_names=False)
    tm.assert_series_equal(out.va_i8, ints, check_names=False)
    tm.assert_series_equal(out.va_i16, ints, check_names=False)
    tm.assert_series_equal(out.va_i32, ints, check_names=False)
    tm.assert_series_equal(out.va_i64, ints, check_names=False)

    floats = pd.Series([[x[0] - 1, x[1] - 2, x[2] - 3] for x in FLOAT_VECTORS])
    ints = pd.Series([[x[0] - 1, x[1] - 2, x[2] - 3] for x in INT_VECTORS])

    tm.assert_series_equal(out.va_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_f32_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_f64_lit, floats, check_names=False)
    tm.assert_series_equal(out.va_i8_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i16_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i32_lit, ints, check_names=False)
    tm.assert_series_equal(out.va_i64_lit, ints, check_names=False)


def test_vector_subvector(con: Any) -> None:
    tbl = make_vectors(con.table('datatypes'))

    out = tbl.mutate(
        vs=lambda x: x.vec1.vector_subvector(1, 2).json_array_unpack(),
        vs_f32=lambda x: x.vec1_f32.vector_subvector_f32(1, 2).json_array_unpack_f32(),
        vs_f64=lambda x: x.vec1_f64.vector_subvector_f64(1, 2).json_array_unpack_f64(),
        vs_i8=lambda x: x.vec1_i8.vector_subvector_i8(1, 2).json_array_unpack_i8(),
        vs_i16=lambda x: x.vec1_i16.vector_subvector_i16(1, 2).json_array_unpack_i16(),
        vs_i32=lambda x: x.vec1_i32.vector_subvector_i32(1, 2).json_array_unpack_i32(),
        vs_i64=lambda x: x.vec1_i64.vector_subvector_i64(1, 2).json_array_unpack_i64(),
    ).order_by('id').execute()

    out = out[out.text_vector.notnull()].reset_index(drop=True)

    floats = pd.Series([
        [2.5, -6.69999981],
        [-3.29999995, 7.80000019],
        [6.5999999, 7.9000001],
    ])
    ints = pd.Series([[3, -7], [-3, 8], [7, 8]])

    tm.assert_series_equal(out.vs, floats, check_names=False)
    tm.assert_series_equal(out.vs_f32, floats, check_names=False)
    tm.assert_series_equal(out.vs_f64, floats, check_names=False)
    tm.assert_series_equal(out.vs_i8, ints, check_names=False)
    tm.assert_series_equal(out.vs_i16, ints, check_names=False)
    tm.assert_series_equal(out.vs_i32, ints, check_names=False)
    tm.assert_series_equal(out.vs_i64, ints, check_names=False)


# def test_vector_sum(con: Any) -> None:
#    tbl = make_vectors(con.table('datatypes'))
#
#    out = tbl.mutate(
#        vs=lambda x: x.vec1.vector_sum(),
#        vs_f32=lambda x: x.vec1_f32.vector_sum_f32(),
#        vs_f64=lambda x: x.vec1_f64.vector_sum_f64(),
#        vs_i8=lambda x: x.vec1_i8.vector_sum_i8(),
#        vs_i16=lambda x: x.vec1_i16.vector_sum_i16(),
#        vs_i32=lambda x: x.vec1_i32.vector_sum_i32(),
#        vs_i64=lambda x: x.vec1_i64.vector_sum_i64(),
#    ).order_by('id').execute()
#
#    out = out[out.text_vector.notnull()].reset_index(drop=True)
#
#    print(out)
#
#    floats = pd.Series([[0.0100000007, 6.25, 44.8899956],
#                        [106.090004, 10.8899994, 60.840004],
#                        [5.28999996, 43.5599976, 62.4099998]])
#    ints = pd.Series([[0, 9, 49], [100, 9, 64], [4, 49, 64]])
#
#    tm.assert_series_equal(out.vs, floats, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_f32, floats, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_f64, floats, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_i8, ints, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_i16, ints, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_i32, ints, check_names=False, check_dtype=False)
#    tm.assert_series_equal(out.vs_i64, ints, check_names=False, check_dtype=False)



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class Hex(Value):
    """Convert bytes to hex values."""

    arg = binary

    output_dtype = dt.string
    output_shape = rlz.shape_like('arg')


def hex(arg: ir.BinaryValue) -> ir.StringValue:
    """
    Convert a collection of bytes into hex values.

    Parameters
    ----------
    arg : binary
        Blob to convert

    Returns
    -------
    String column

    """
    return Hex(arg).to_expr()


ir.BinaryValue.hex = hex


class Unhex(Value):
    """Convert hex string to bytes."""

    arg = rlz.string

    output_dtype = dt.binary
    output_shape = rlz.shape_like('arg')


def unhex(arg: ir.StringValue) -> ir.BinaryValue:
    """
    Convert a string of hex values to bytes.

    Parameters
    ----------
    arg : string
        Hex string to convert

    Returns
    -------
    Binary column

    """
    return Unhex(arg).to_expr()


ir.StringValue.unhex = unhex



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class ScalarVectorMul(Value):
    """32-bit float dot product."""

    arg = binary
    n = rlz.strict_numeric

    output_dtype = dt.binary
    output_shape = rlz.shape_like('arg')


def scalar_vector_mul(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 32-bit float vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMul(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul = scalar_vector_mul


class ScalarVectorMulI8(ScalarVectorMul):
    """8-bit integer dot product."""


def scalar_vector_mul_i8(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 8-bit int vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulI8(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_i8 = scalar_vector_mul_i8


class ScalarVectorMulI16(ScalarVectorMul):
    """16-bit integer dot product."""


def scalar_vector_mul_i16(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 16-bit int vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulI16(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_i16 = scalar_vector_mul_i16


class ScalarVectorMulI32(ScalarVectorMul):
    """32-bit integer dot product."""


def scalar_vector_mul_i32(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 32-bit int vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulI32(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_i32 = scalar_vector_mul_i32


class ScalarVectorMulI64(ScalarVectorMul):
    """64-bit integer dot product."""


def scalar_vector_mul_i64(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 64-bit int vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulI64(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_i64 = scalar_vector_mul_i64


class ScalarVectorMulF32(ScalarVectorMul):
    """32-bit float dot product."""


def scalar_vector_mul_f32(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 32-bit float vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulF32(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_f32 = scalar_vector_mul_f32


class ScalarVectorMulF64(ScalarVectorMul):
    """64-bit float dot product."""


def scalar_vector_mul_f64(arg: ir.BinaryValue, n: ir.NumericValue) -> ir.BinaryValue:
    """
    Multiply each element in the 64-bit float vector blob by the specified value.

    Parameters
    ----------
    arg : blob
        Vector expression
    n : int or float
        Multiplier value

    Returns
    -------
    Blob column

    """
    return ScalarVectorMulF64(arg, n).to_expr()


ir.BinaryValue.scalar_vector_mul_f64 = scalar_vector_mul_f64



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations import Reduction

binary = rlz.value(dt.binary)


class VectorSum(Reduction):
    """32-bit float columnar vector sum."""

    arg = binary

    output_dtype = dt.binary


def vector_sum(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 32-bit float vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSum(arg).to_expr()


ir.BinaryValue.vector_sum = vector_sum


class VectorSumI8(VectorSum):
    """8-bit integer columnar vector sum."""


def vector_sum_i8(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 8-bit int vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSumI8(arg).to_expr()


ir.BinaryValue.vector_sum_i8 = vector_sum_i8


class VectorSumI16(VectorSum):
    """16-bit integer columnar vector sum."""


def vector_sum_i16(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 16-bit int vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSum(arg).to_expr()


ir.BinaryValue.vector_sum_i16 = vector_sum_i16


class VectorSumI32(VectorSum):
    """32-bit integer columnar vector sum."""


def vector_sum_i32(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 32-bit int vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSumI32(arg).to_expr()


ir.BinaryValue.vector_sum_i32 = vector_sum_i32


class VectorSumI64(VectorSum):
    """64-bit integer columnar vector sum."""


def vector_sum_i64(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 64-bit int vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSumI64(arg).to_expr()


ir.BinaryValue.vector_sum_i64 = vector_sum_i64


class VectorSumF32(VectorSum):
    """32-bit float columnar vector sum."""


def vector_sum_f32(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 32-bit float vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSumF32(arg).to_expr()


ir.BinaryValue.vector_sum_f32 = vector_sum_f32


class VectorSumF64(VectorSum):
    """64-bit float columnar vector sum."""


def vector_sum_f64(arg: ir.BinaryValue) -> ir.BinaryValue:
    """
    Return a columnar sum of 64-bit float vectors.

    Parameters
    ----------
    arg : blob
        Vector expression

    Returns
    -------
    Blob column

    """
    return VectorSumF64(arg).to_expr()


ir.BinaryValue.vector_sum_f64 = vector_sum_f64



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONGetType(Value):
    """JSON type getter."""

    arg = rlz.one_of([rlz.json, rlz.string])

    output_dtype = dt.string
    output_shape = rlz.shape_like('arg')


def json_get_type(arg: ir.JSONValue | ir.StringValue) -> ir.StringValue:
    """
    Get the type of the JSON object.

    Parameters
    ----------
    arg : JSON or string
        JSON array

    Returns
    -------
    String column

    """
    return JSONGetType(arg).to_expr()


ir.JSONValue.json_get_type = json_get_type
ir.StringValue.json_get_type = json_get_type



from __future__ import annotations

import json
from typing import Any

import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONSpliceDouble(Value):
    """JSON splice double values."""

    arg = rlz.one_of([rlz.json, rlz.string])
    start = rlz.integer
    length = rlz.integer
    values = rlz.tuple_of(rlz.strict_numeric, min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_splice_double(
    arg: ir.JSONValue | ir.StringValue,
    start: ir.IntegerValue,
    length: ir.IntegerValue,
    *values: ir.FloatingValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Splice double values at the specified position.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    start : int
        Zero-indexed position to start splice
    length : int
        Length of the splice
    *values : doubles
        Values to insert

    Returns
    -------
    String or JSON Column

    """
    return JSONSpliceDouble(arg, start, length, values).to_expr()


ir.JSONValue.json_splice_double = json_splice_double
ir.StringValue.json_splice_double = json_splice_double


class JSONSpliceString(Value):
    """JSON splice string values."""

    arg = rlz.one_of([rlz.json, rlz.string])
    start = rlz.integer
    length = rlz.integer
    values = rlz.tuple_of(rlz.string, min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_splice_string(
    arg: ir.JSONValue | ir.StringValue,
    start: ir.IntegerValue,
    length: ir.IntegerValue,
    *values: ir.StringValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Splice double values at the specified position.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    start : int
        Zero-indexed position to start splice
    length : int
        Length of the splice
    *values : strings
        Values to insert

    Returns
    -------
    String or JSON Column

    """
    return JSONSpliceString(arg, start, length, values).to_expr()


ir.JSONValue.json_splice_string = json_splice_string
ir.StringValue.json_splice_string = json_splice_string


class JSONSpliceJSON(Value):
    """JSON splice JSON values."""

    arg = rlz.one_of([rlz.json, rlz.string])
    start = rlz.integer
    length = rlz.integer
    values = rlz.tuple_of(rlz.one_of([rlz.json, rlz.string]), min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_splice_json(
    arg: ir.JSONValue | ir.StringValue,
    start: ir.IntegerValue,
    length: ir.IntegerValue,
    *values: ir.JSONValue | ir.StringValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Splice JSON values at the specified position.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    start : int
        Zero-indexed position to start splice
    length : int
        Length of the splice
    *values : strings
        Values to insert

    Returns
    -------
    String or JSON Column

    """
    return JSONSpliceJSON(arg, start, length, values).to_expr()


ir.JSONValue.json_splice_json = json_splice_json
ir.StringValue.json_splice_json = json_splice_json


class JSONSplice(Value):
    """JSON splice values."""

    arg = rlz.one_of([rlz.json, rlz.string])
    start = rlz.integer
    length = rlz.integer
    values = rlz.tuple_of(rlz.any, min_length=1)

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_splice(
    arg: ir.JSONValue | ir.StringValue,
    start: ir.IntegerValue,
    length: ir.IntegerValue,
    *values: ir.Value | Any,
) -> ir.JSONValue | ir.StringValue:
    """
    Splice JSON values at the specified position.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    start : int
        Zero-indexed position to start splice
    length : int
        Length of the splice
    *values : Any
        Values to insert

    Returns
    -------
    String or JSON Column

    """
    items = tuple(values)
    if values and not isinstance(values[0], ir.Value):
        items = tuple([json.dumps(x) for x in items])
    return JSONSpliceJSON(arg, start, length, items).to_expr()


ir.JSONValue.json_splice = json_splice
ir.StringValue.json_splice = json_splice



from __future__ import annotations

import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONKeys(Value):
    """JSON exclude using a mask."""

    arg = rlz.one_of([rlz.json, rlz.string])
    key_path = rlz.tuple_of(rlz.one_of([rlz.string, rlz.integer]))

    output_dtype = dt.json
    output_shape = rlz.shape_like('arg')


def json_keys(
    arg: ir.JSONValue | ir.StringValue, *key_path: ir.StringValue | ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Return the top-level keys of a JSON object in the form of a JSON array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    *key_path : strings or ints
        Path to a JSON object

    Returns
    -------
    JSON or String column

    """
    return JSONKeys(arg, key_path).to_expr()


ir.JSONValue.json_keys = json_keys
ir.StringValue.json_keys = json_keys



from __future__ import annotations

import sqlalchemy as sa
from ibis.backends.base.sql.alchemy import AlchemyCompiler
from ibis.backends.base.sql.alchemy import AlchemyExprTranslator

from .datatypes import SingleStoreDBType
from .registry import operation_registry


class SingleStoreDBExprTranslator(AlchemyExprTranslator):
    _registry = operation_registry.copy()
    _rewrites = AlchemyExprTranslator._rewrites.copy()
    _integer_to_timestamp = sa.func.from_unixtime
    native_json_type = False
    _dialect_name = 'singlestoredb'
    type_mapper = SingleStoreDBType


rewrites = SingleStoreDBExprTranslator.rewrites


class SingleStoreDBCompiler(AlchemyCompiler):
    translator_class = SingleStoreDBExprTranslator
    support_values_syntax_in_select = False



from __future__ import annotations

import json
from typing import Any

import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONArrayPushDouble(Value):
    """JSON array append."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.strict_numeric

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_array_push_double(
    arg: ir.JSONValue | ir.StringValue, value: ir.NumericValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Append a double to the end of the JSON array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : float
        Float value to append

    Returns
    -------
    JSON or string column

    """
    return JSONArrayPushDouble(arg, value).to_expr()


ir.JSONValue.json_array_push_double = json_array_push_double
ir.StringValue.json_array_push_double = json_array_push_double


class JSONArrayPushString(Value):
    """JSON array append."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.string

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_array_push_string(
    arg: ir.JSONValue | ir.StringValue, value: ir.StringValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Append a string to the end of the JSON array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : string
        String value to append

    Returns
    -------
    JSON or string column

    """
    return JSONArrayPushString(arg, value).to_expr()


ir.JSONValue.json_array_push_string = json_array_push_string
ir.StringValue.json_array_push_string = json_array_push_string


class JSONArrayPushJSON(Value):
    """JSON array append."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.one_of([rlz.json, rlz.string])

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_array_push_json(
    arg: ir.JSONValue | ir.StringValue, value: ir.JSONValue | ir.StringValue,
) -> ir.JSONValue | ir.StringValue:
    """
    Append a JSON object to the end of the JSON array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : JSON or string
        JSON value to append

    Returns
    -------
    JSON or string column

    """
    return JSONArrayPushJSON(arg, value).to_expr()


ir.JSONValue.json_array_push_json = json_array_push_json
ir.StringValue.json_array_push_json = json_array_push_json


class JSONArrayPush(Value):
    """JSON array append."""

    arg = rlz.one_of([rlz.json, rlz.string])
    value = rlz.any

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_array_push(
    arg: ir.JSONValue | ir.StringValue, value: ir.Value | Any,
) -> ir.JSONValue | ir.StringValue:
    """
    Append a JSON object to the end of the JSON array.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    value : Any
        Value to append

    Returns
    -------
    JSON or string column

    """
    if not isinstance(value, ir.Value):
        value = json.dumps(value)
    return JSONArrayPushJSON(arg, value).to_expr()


ir.JSONValue.json_array_push = json_array_push
ir.StringValue.json_array_push = json_array_push



from __future__ import annotations

from .json_agg import *
from .json_array_contains import *
from .json_array_push import *
from .json_delete_key import *
from .json_extract import *
from .json_get_type import *
from .json_include_exclude_mask import *
from .json_keys import *
from .json_length import *
from .json_match_any import *
from .json_pretty import *
from .json_set import *
from .json_splice import *
from .json_to_array import *
from .to_json import *



from __future__ import annotations

import json
from typing import Any

import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value


class JSONExcludeMask(Value):
    """JSON exclude using a mask."""

    arg = rlz.one_of([rlz.json, rlz.string])
    mask = rlz.string

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_exclude_mask(
    arg: ir.JSONValue | ir.StringValue,
    mask: ir.JSONValue | ir.StringValue | dict[str, Any],
) -> ir.JSONValue | ir.StringValue:
    """
    Get a subset of a JSON object after applying a mask.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    mask : JSON or string or dict
        Mask to apply

    Returns
    -------
    JSON or String column

    """
    if isinstance(mask, dict):
        mask = json.dumps(mask)
    return JSONExcludeMask(arg, mask).to_expr()


ir.JSONValue.json_exclude_mask = json_exclude_mask
ir.StringValue.json_exclude_mask = json_exclude_mask


class JSONIncludeMask(Value):
    """JSON include using a mask."""

    arg = rlz.one_of([rlz.json, rlz.string])
    mask = rlz.string

    output_dtype = rlz.dtype_like('arg')
    output_shape = rlz.shape_like('arg')


def json_include_mask(
    arg: ir.JSONValue | ir.StringValue,
    mask: ir.JSONValue | ir.StringValue | dict[str, Any],
) -> ir.JSONValue | ir.StringValue:
    """
    Get a subset of a JSON object after applying a mask.

    Parameters
    ----------
    arg : JSON or string
        JSON array
    mask : JSON or string or dict
        Mask to apply

    Returns
    -------
    JSON or String column

    """
    if isinstance(mask, dict):
        mask = json.dumps(mask)
    return JSONIncludeMask(arg, mask).to_expr()


ir.JSONValue.json_include_mask = json_include_mask
ir.StringValue.json_include_mask = json_include_mask



from __future__ import annotations

import json
from typing import Iterable

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr.operations.core import Value

binary = rlz.value(dt.binary)


class VectorAdd(Value):
    """32-bit float vector addition."""

    left = binary
    right = binary

    output_dtype = dt.binary
    output_shape = rlz.shape_like('left')


def vector_add(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 32-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack()
    return VectorAdd(left, right).to_expr()


ir.BinaryValue.vector_add = vector_add


class VectorAddI8(VectorAdd):
    """8-bit integer vector addition."""


def vector_add_i8(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 8-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i8()
    return VectorAddI8(left, right).to_expr()


ir.BinaryValue.vector_add_i8 = vector_add_i8


class VectorAddI16(VectorAdd):
    """16-bit integer vector addition."""


def vector_add_i16(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 16-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i16()
    return VectorAddI16(left, right).to_expr()


ir.BinaryValue.vector_add_i16 = vector_add_i16


class VectorAddI32(VectorAdd):
    """32-bit integer vector addition."""


def vector_add_i32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 32-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i32()
    return VectorAddI32(left, right).to_expr()


ir.BinaryValue.vector_add_i32 = vector_add_i32


class VectorAddI64(VectorAdd):
    """64-bit integer vector addition."""


def vector_add_i64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 64-bit int input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_i64()
    return VectorAddI64(left, right).to_expr()


ir.BinaryValue.vector_add_i64 = vector_add_i64


class VectorAddF32(VectorAdd):
    """32-bit float vector addition."""


def vector_add_f32(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 32-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f32()
    return VectorAddF32(left, right).to_expr()


ir.BinaryValue.vector_add_f32 = vector_add_f32


class VectorAddF64(VectorAdd):
    """64-bit float vector addition."""


def vector_add_f64(
    left: ir.BinaryValue,
    right: ir.BinaryValue | Iterable[float | int],
) -> ir.BinaryValue:
    """
    Add two 64-bit float input vectors.

    Parameters
    ----------
    left : blob
        Vector expression
    right : blob or iterable
        Vector expression

    Returns
    -------
    Blob column

    """
    if not isinstance(right, ir.Value):
        right = ibis.literal(json.dumps(list(right))).json_array_pack_f64()
    return VectorAddF64(left, right).to_expr()


ir.BinaryValue.vector_add_f64 = vector_add_f64
